@@ -12,7 +12,7 @@ from typing import Any, NoReturn
 from typing import TypeVar, AnyStr
 from typing import T, KT, VT  # Not in __all__.
 from typing import Union, Optional
-from typing import Tuple, List, MutableMapping
+from typing import Tuple, List, MutableMapping, Iterator
 from typing import Callable
 from typing import Generic, ClassVar, GenericMeta
 from typing import cast
@@ -1497,6 +1497,65 @@ class ForwardRefTests(BaseTestCase):
         self.assertEqual(fr, typing._ForwardRef('int'))
         self.assertNotEqual(List['int'], List[int])
 
+    def test_forward_equality_gth(self):
+        c1 = typing._ForwardRef('C')
+        c1_gth = typing._ForwardRef('C')
+        c2 = typing._ForwardRef('C')
+        c2_gth = typing._ForwardRef('C')
+
+        class C:
+            pass
+        def foo(a: c1_gth, b: c2_gth):
+            pass
+
+        self.assertEqual(get_type_hints(foo, globals(), locals()), {'a': C, 'b': C})
+        self.assertEqual(c1, c2)
+        self.assertEqual(c1, c1_gth)
+        self.assertEqual(c1_gth, c2_gth)
+        self.assertEqual(List[c1], List[c1_gth])
+        self.assertNotEqual(List[c1], List[C])
+        self.assertNotEqual(List[c1_gth], List[C])
+        self.assertEquals(Union[c1, c1_gth], Union[c1])
+        self.assertEquals(Union[c1, c1_gth, int], Union[c1, int])
+
+    def test_forward_equality_hash(self):
+        c1 = typing._ForwardRef('int')
+        c1_gth = typing._ForwardRef('int')
+        c2 = typing._ForwardRef('int')
+        c2_gth = typing._ForwardRef('int')
+
+        def foo(a: c1_gth, b: c2_gth):
+            pass
+        get_type_hints(foo, globals(), locals())
+
+        self.assertEqual(hash(c1), hash(c2))
+        self.assertEqual(hash(c1_gth), hash(c2_gth))
+        self.assertEqual(hash(c1), hash(c1_gth))
+
+    def test_forward_equality_namespace(self):
+        class A:
+            pass
+        def namespace1():
+            a = typing._ForwardRef('A')
+            def fun(x: a):
+                pass
+            get_type_hints(fun, globals(), locals())
+            return a
+
+        def namespace2():
+            a = typing._ForwardRef('A')
+
+            class A:
+                pass
+            def fun(x: a):
+                pass
+
+            get_type_hints(fun, globals(), locals())
+            return a
+
+        self.assertEqual(namespace1(), namespace1())
+        self.assertNotEqual(namespace1(), namespace2())
+
     def test_forward_repr(self):
         self.assertEqual(repr(List['int']), "typing.List[_ForwardRef('int')]")
 
@@ -1516,6 +1575,67 @@ class ForwardRefTests(BaseTestCase):
         self.assertEqual(get_type_hints(foo, globals(), locals()),
                          {'a': Tuple[T]})
 
+    def test_forward_recursion_actually(self):
+        def namespace1():
+            a = typing._ForwardRef('A')
+            A = a
+            def fun(x: a): pass
+
+            ret = get_type_hints(fun, globals(), locals())
+            return a
+
+        def namespace2():
+            a = typing._ForwardRef('A')
+            A = a
+            def fun(x: a): pass
+
+            ret = get_type_hints(fun, globals(), locals())
+            return a
+
+        def cmp(o1, o2):
+            return o1 == o2
+
+        r1 = namespace1()
+        r2 = namespace2()
+        self.assertIsNot(r1, r2)
+
+        try:
+            exc = RecursionError
+        except NameError:
+            exc = RuntimeError
+        self.assertRaises(exc, cmp, r1, r2)
+
+    def test_union_forward_recursion(self):
+        ValueList = List['Value']
+        Value = Union[str, ValueList]
+
+        def c(foo: List[Value]):
+            pass
+        def d(foo: Union[Value, ValueList]):
+            pass
+        def e(foo: Union[List[Value], ValueList]):
+            pass
+        def f(foo: Union[Value, List[Value], ValueList]):
+            pass
+
+        self.assertEqual(get_type_hints(c, globals(), locals()),
+                         get_type_hints(c, globals(), locals()))
+        self.assertEqual(get_type_hints(c, globals(), locals()),
+                         {'foo': List[Union[str, List[Union[str, List['Value']]]]]})
+        self.assertEqual(get_type_hints(d, globals(), locals()),
+                         {'foo': Union[str, List[Union[str, List['Value']]]]})
+        self.assertEqual(get_type_hints(e, globals(), locals()),
+                         {'foo': Union[
+                             List[Union[str, List[Union[str, List['Value']]]]],
+                             List[Union[str, List['Value']]]
+                         ]})
+        self.assertEqual(get_type_hints(f, globals(), locals()),
+                         {'foo': Union[
+                             str,
+                             List[Union[str, List['Value']]],
+                             List[Union[str, List[Union[str, List['Value']]]]]
+                         ]})
+
     def test_callable_forward(self):
 
         def foo(a: Callable[['T'], 'T']):
@@ -1539,7 +1659,7 @@ class ForwardRefTests(BaseTestCase):
 
     def test_delayed_syntax_error(self):
 
-        def foo(a: 'Node[T'):
+        def foo(a: 'Node[T'):  # noqa
             pass
 
         with self.assertRaises(SyntaxError):
@@ -1555,7 +1675,7 @@ class ForwardRefTests(BaseTestCase):
 
     def test_name_error(self):
 
-        def foo(a: 'Noode[T]'):
+        def foo(a: 'Noode[T]'):  # noqa
             pass
 
         with self.assertRaises(NameError):
@@ -1564,7 +1684,7 @@ class ForwardRefTests(BaseTestCase):
     def test_no_type_check(self):
 
         @no_type_check
-        def foo(a: 'whatevers') -> {}:
+        def foo(a: 'whatevers') -> {}:  # noqa
             pass
 
         th = get_type_hints(foo)
@@ -1574,7 +1694,7 @@ class ForwardRefTests(BaseTestCase):
 
         @no_type_check
         class C:
-            def foo(a: 'whatevers') -> {}:
+            def foo(a: 'whatevers') -> {}:  # noqa
                 pass
 
         cth = get_type_hints(C.foo)
@@ -1600,12 +1720,12 @@ class ForwardRefTests(BaseTestCase):
         self.assertEqual(magic_decorator.__name__, 'magic_decorator')
 
         @magic_decorator
-        def foo(a: 'whatevers') -> {}:
+        def foo(a: 'whatevers') -> {}:  # noqa
             pass
 
         @magic_decorator
         class C:
-            def foo(a: 'whatevers') -> {}:
+            def foo(a: 'whatevers') -> {}:  # noqa
                 pass
 
         self.assertEqual(foo.__name__, 'foo')
@@ -1764,7 +1884,7 @@ else:
     # fake names for the sake of static analysis
     ann_module = ann_module2 = ann_module3 = None
     A = B = CSub = G = CoolEmployee = CoolEmployeeWithDefault = object
-    XMeth = XRepr = NoneAndForward = object
+    XMeth = XRepr = NoneAndForward = HasForeignBaseClass = object
 
 gth = get_type_hints
 
@@ -1826,7 +1946,7 @@ class GetTypeHintTests(BaseTestCase):
         @no_type_check
         class NoTpCheck:
             class Inn:
-                def __init__(self, x: 'not a type'): ...
+                def __init__(self, x: 'not a type'): ...  # noqa
         self.assertTrue(NoTpCheck.__no_type_check__)
         self.assertTrue(NoTpCheck.Inn.__init__.__no_type_check__)
         self.assertEqual(gth(ann_module2.NTC.meth), {})
@@ -2233,8 +2353,8 @@ class CollectionsAbcTests(BaseTestCase):
         self.assertIsSubclass(MMB[str, str], typing.Mapping)
         self.assertIsSubclass(MMC, MMA)
 
-        class I(typing.Iterable): ...
-        self.assertNotIsSubclass(list, I)
+        class It(typing.Iterable): ...
+        self.assertNotIsSubclass(list, It)
 
         class G(typing.Generator[int, int, int]): ...
         def g(): yield 0
@@ -2316,8 +2436,8 @@ class CollectionsAbcTests(BaseTestCase):
         self.assertIsSubclass(S, typing.MutableSequence)
         self.assertIsSubclass(S, typing.Iterable)
 
-        class I(collections_abc.Iterable): ...
-        self.assertIsSubclass(I, typing.Iterable)
+        class It(collections_abc.Iterable): ...
+        self.assertIsSubclass(It, typing.Iterable)
 
         class A(collections_abc.Mapping, metaclass=abc.ABCMeta): ...
         class B: ...
@@ -2513,6 +2633,38 @@ class XMethBad2(NamedTuple):
         with self.assertRaises(TypeError):
             NamedTuple('Name', x=1, y='a')
 
+    @skipUnless(PY36, 'Python 3.6 required')
+    def test_namedtuple_special_keyword_names(self):
+        NT = NamedTuple("NT", cls=type, self=object, typename=str, fields=list)
+        self.assertEqual(NT.__name__, 'NT')
+        self.assertEqual(NT._fields, ('cls', 'self', 'typename', 'fields'))
+        a = NT(cls=str, self=42, typename='foo', fields=[('bar', tuple)])
+        self.assertEqual(a.cls, str)
+        self.assertEqual(a.self, 42)
+        self.assertEqual(a.typename, 'foo')
+        self.assertEqual(a.fields, [('bar', tuple)])
+
+    @skipUnless(PY36, 'Python 3.6 required')
+    def test_namedtuple_errors(self):
+        with self.assertRaises(TypeError):
+            NamedTuple.__new__()
+        with self.assertRaises(TypeError):
+            NamedTuple()
+        with self.assertRaises(TypeError):
+            NamedTuple('Emp', [('name', str)], None)
+        with self.assertRaises(ValueError):
+            NamedTuple('Emp', [('_name', str)])
+
+        with self.assertWarns(DeprecationWarning):
+            Emp = NamedTuple(typename='Emp', name=str, id=int)
+        self.assertEqual(Emp.__name__, 'Emp')
+        self.assertEqual(Emp._fields, ('name', 'id'))
+
+        with self.assertWarns(DeprecationWarning):
+            Emp = NamedTuple('Emp', fields=[('name', str), ('id', int)])
+        self.assertEqual(Emp.__name__, 'Emp')
+        self.assertEqual(Emp._fields, ('name', 'id'))
+
     def test_pickle(self):
         global Emp  # pickle wants to reference the class by name
         Emp = NamedTuple('Emp', [('name', str), ('id', int)])
@@ -2653,9 +2805,8 @@ class AllTests(BaseTestCase):
         file_path = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                  'typing.py')
         try:
-            subprocess.check_output('python -OO {}'.format(file_path),
-                                    stderr=subprocess.STDOUT,
-                                    shell=True)
+            subprocess.check_output([sys.executable, '-OO', file_path],
+                                    stderr=subprocess.STDOUT)
         except subprocess.CalledProcessError:
             self.fail('Module does not compile with optimize=2 (-OO flag).')
 
